import argparse
import os
import signal
from pathlib import Path
from typing import Dict, Tuple

os.environ["PYQTGRAPH_QT_LIB"] = "PyQt6"
from importlib.resources import files
from importlib.resources import path as resource_path

import pyvisa
import qdarktheme
from PyQt6.QtCore import QLocale
from PyQt6.QtGui import QGuiApplication, QIcon
from PyQt6.QtWidgets import QApplication, QStackedWidget, QWidget

from sonaris.defaults import (
    MONITOR_FILE,
    OSCILLOSCOPE_BUFFER_SIZE,
    TIMEKEEPER_JOBS_FILE,
    DeviceName,
    APP_NAME,
    VERSION_STRING,
    DEFAULT_DATADIR,
    GF_DATASOURCES_DIR,
    GF_DASHBOARDS_DIR,
)
from sonaris.frontend.managers.dg4202 import DG4202Manager
from sonaris.frontend.managers.edux1002a import EDUX1002AManager
from sonaris.frontend.managers.state_manager import StateManager
from sonaris import factory
from sonaris.frontend.pages.general import GeneralPage
from sonaris.frontend.pages.monitor import MonitorPage
from sonaris.frontend.pages.scheduler import SchedulerPage
from sonaris.frontend.pages.settings import SettingsPage
from sonaris.frontend.tasks.tasks import get_tasks
from sonaris.frontend.widgets.menu import MainMenuBar
from sonaris.frontend.widgets.sidebar import Sidebar
from sonaris.frontend.widgets.templates import ModularMainWindow
from sonaris.scheduler import registry
from sonaris.scheduler.timekeeper import Timekeeper
from sonaris.scheduler.worker import Worker
from sonaris.utils.log import get_logger
from sonaris.services.grafana import GrafanaContainerService
from sonaris.services.datasource import DataSourceService
logger = get_logger()

# Before creating your application instance
QLocale.setDefault(QLocale(QLocale.Language.English, QLocale.Country.UnitedStates))

logger.info(f"{APP_NAME} {VERSION_STRING}")
logger.info(f"Using {DEFAULT_DATADIR} as working directory.")
logger.info(f"Using {TIMEKEEPER_JOBS_FILE} as persistence file.")
logger.info(f"Using {OSCILLOSCOPE_BUFFER_SIZE} oscilloscope buffer size.")
logger.info(f"Device events under {MONITOR_FILE}.")

def signal_handler(signum, frame):
    logger.info("Exit signal detected.")
    QApplication.quit()
    factory.grafana_container_service.stop()
    factory.data_source_service.stop()
    factory.worker.stop_worker()
    # Invoke the default SIGINT handler to exit the application
    signal.signal(signum, signal.SIG_DFL)
    os.kill(os.getpid(), signum)


def init_objects(args_dict: dict):
    #================= Hardware Managers===================#
    factory.resource_manager = pyvisa.ResourceManager()
    factory.state_manager = StateManager()
    factory.edux1002a_manager = EDUX1002AManager(
        state_manager=factory.state_manager,
        args_dict=args_dict,
        resource_manager=factory.resource_manager,
        buffer_size=OSCILLOSCOPE_BUFFER_SIZE,
    )
    factory.dg4202_manager = DG4202Manager(
        factory.state_manager,
        args_dict=args_dict,
        resource_manager=factory.resource_manager,
    )
    factory.worker = Worker(
        function_map=registry.function_map,
        logger=logger,
    )
    factory.timekeeper = Timekeeper(
        persistence_file=TIMEKEEPER_JOBS_FILE,
        worker_instance=factory.worker,
        logger=logger,
    )

    #================= Register Tasks ===================#
    for task_name, func_pointer in get_tasks(flatten=True).items():
        factory.worker.register_task(func_pointer, task_name)
    #==================== Services ======================#
    if args_dict["grafana"]:
        factory.grafana_container_service = GrafanaContainerService(
            client=None, # Use default client instance
            port=None # Use default port
        )
        factory.data_source_service = DataSourceService(
            timekeeper=factory.timekeeper,
            port=None, # Use default port
            logger=logger,
            name=f"{APP_NAME}DataSource" # Customize as needed
            )
        factory.data_source_service.write_provisioning_files(dashboards_dir=GF_DASHBOARDS_DIR, datasources_dir=GF_DATASOURCES_DIR)
        factory.data_source_service.start()
        factory.grafana_container_service.start()
    factory.worker.start_worker()


class MainWindow(ModularMainWindow):
    def __init__(self, args_dict: dict) -> None:
        super().__init__()
        menu_bar = MainMenuBar(self)
        self.setWindowTitle(APP_NAME)
        self.setMenuBar(menu_bar)
        self.last_page = ""
        # ---------------------------------------------------------------------- #
        # ---------------------------SIDEBAR SETUP------------------------------ #
        self.sidebar = Sidebar(self)
        device_managers = {
            DeviceName.DG4202.value: factory.dg4202_manager,
            DeviceName.EDUX1002A.value: factory.edux1002a_manager,
        }
        self.sidebar.sizePolicy()
        self.sidebar_dict: Dict[str, QWidget] = {
            GeneralPage.PAGE_NAME: GeneralPage(
                dg4202_manager=factory.dg4202_manager,
                edux1002a_manager=factory.edux1002a_manager,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
            SchedulerPage.PAGE_NAME: SchedulerPage(
                timekeeper=factory.timekeeper,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
            MonitorPage.PAGE_NAME: MonitorPage(
                device_managers=device_managers,
                parent=self,
                args_dict=args_dict,
                monitor_logs=MONITOR_FILE,
                root_callback=self.root_callback,
            ),
            SettingsPage.PAGE_NAME: SettingsPage(
                device_managers=device_managers,
                parent=self,
                args_dict=args_dict,
                root_callback=self.root_callback,
            ),
        }
        self.sidebar.addItems(self.sidebar_dict.keys())  # Add strings to sidebar items
        self.sidebar_content = QStackedWidget(self)
        list(
            map(self.sidebar_content.addWidget, self.sidebar_dict.values())
        )  # Adds all child widgets to content widgets

        # --------------------------------------------------------------------- #
        # ---------------------------LAYOUT SETUP------------------------------ #
        # --------------------------------------------------------------------- #

        self.add_widget_to_left(self.sidebar)
        self.add_widget_to_middle(
            self.sidebar_content
        )  # Use the method from ModularMainWindow
        # Connect the Sidebar's custom signal to the MainWindow's slot
        self.sidebar.pageSelected.connect(self.loadPage)

    def loadPage(self, page_name: str) -> None:
        page_widget: QWidget = self.sidebar_dict.get(page_name)
        if page_widget and self.last_page != page_name:
            self.sidebar_content.setCurrentWidget(page_widget)
            # logger.info(f"[Sidebar] Switch event to {page_name}")
            page_widget.update()
            self.last_page = page_name

    def root_callback(self):
        """tells pages the pages to call update method."""
        for _, page_obj in self.sidebar_dict.items():
            page_obj.update()

    def closeEvent(self, event):
        logger.info("main x exit button was clicked")
        factory.worker.stop_worker()
        super().closeEvent(event)

    def showEvent(self, event):
        primaryScreen = QGuiApplication.primaryScreen()
        screenGeometry = primaryScreen.availableGeometry()
        centerPoint = screenGeometry.center()
        frameGeometry = self.frameGeometry()
        frameGeometry.moveCenter(centerPoint)
        self.move(frameGeometry.topLeft())
        super().showEvent(event)


def create_app(args_dict: dict) -> Tuple[QApplication, MainWindow]:
    init_objects(args_dict=args_dict)

    app = QApplication([])
    window = MainWindow(args_dict)

    qdarktheme.setup_theme()

    # Assuming 'sonaris.assets' is a valid package path
    favicon_path = files("sonaris.assets").joinpath("favicon.ico")
    app_icon = QIcon(str(favicon_path))

    app.setWindowIcon(app_icon)
    window.setWindowIcon(app_icon)

    window.setWindowTitle(APP_NAME)
    window.resize(640, 400)
    logger.info(f"Window size after resize: {window.size()} ")
    return app, window


def run_application(args_dict):
    logger.info(args_dict)
    app, window = create_app(args_dict)
    window.show()
    app.exec()


if __name__ == "__main__":
    # Not recommended anymore to run this way. (sonaris/__main__.py controls the argument parsing using click)
    # Do python -m sonaris run --hardware-mock to run the app in hardware mock mode instead ( or poetry run python -m sonaris run --hardware-mock)
    # You can do packaged runs as well by doing 'pip install .'
    parser = argparse.ArgumentParser(description="Run the sonaris application.")
    parser.add_argument(
        "--hardware-mock",
        action="store_true",
        help="Run the app in hardware mock mode.",
    )
    args = parser.parse_args()
    args_dict = vars(args)
    run_application(args_dict)



from sonaris.utils.log import get_logger
from fastapi import FastAPI
from pathlib import Path
from typing import Optional
from sonaris.scheduler.timekeeper import Timekeeper
from logging import Logger
from sonaris.defaults import DATA_SOURCE_PORT, DATA_SOURCE_NAME, GF_DASHBOARDS_DIR, GF_DATASOURCES_DIR

from sonaris.services.service import MultithreadedServer,Service
import yaml
import json

logger = get_logger()

class DataSourceService(Service):
    def __init__(self, timekeeper: Timekeeper, port: int = None, logger: Optional[Logger] = None, name:str = None):
        super().__init__()
        self.name = str(name) or str(DATA_SOURCE_NAME)
        self.timekeeper = timekeeper
        self.port = port or DATA_SOURCE_PORT
        self.logger = logger or get_logger()
        self.app = FastAPI(title="Sonaris Data Source Service")
        self.setup_routes()

        # Prepare the Uvicorn config here
        self.server = MultithreadedServer(app=self.app, port=self.port, log_level="info")

    @staticmethod
    def get_data_source_manifest(host="localhost", port=DATA_SOURCE_PORT,name=DATA_SOURCE_NAME):
        return {
            "apiVersion": 1,
            "datasources": [
                {
                    "name": name,
                    "type": "grafana-simple-json-datasource",
                    "access": "proxy",
                    "url": f"http://{host}:{port}",
                    "isDefault": True,
                    "jsonData": {},
                    "secureJsonFields": {}
                }
            ]
        }
    @staticmethod
    def get_dashboard_manifest():
        return {
            "dashboard": {
                "id": None,
                "uid": None,
                "title": "Sonaris Data",
                "panels": [
                    {
                        "type": "table",
                        "title": "Archive",
                        "datasource": "SimpleDataSource",
                        "targets": [
                            {
                                "target": "archive",
                                "refId": "A"
                            }
                        ],
                        "transformations": [
                            {
                                "id": "json",
                                "options": {}
                            }
                        ]
                    },
                    {
                        "type": "table",
                        "title": "Jobs",
                        "datasource": "SimpleDataSource",
                        "targets": [
                            {
                                "target": "jobs",
                                "refId": "B"
                            }
                        ],
                        "transformations": [
                            {
                                "id": "json",
                                "options": {}
                            }
                        ]
                    }
                ]
            },
            "overwrite": False
        }
    def write_provisioning_files(self, datasources_dir: Path =  None, dashboards_dir: Path = None):
        # Ensure directories exist
        datasources_dir = datasources_dir or GF_DATASOURCES_DIR
        dashboards_dir = dashboards_dir or GF_DASHBOARDS_DIR
        
        datasources_dir.mkdir(parents=True, exist_ok=True)
        dashboards_dir.mkdir(parents=True, exist_ok=True)

        # Data source file
        data_source_path = datasources_dir / 'datasource.yml'
        with open(data_source_path, 'w') as file:
            yaml.dump(self.get_data_source_manifest(), file)

        # Dashboard file
        dashboard_manifest = self.get_dashboard_manifest()
        dashboard_path = dashboards_dir / f'{self.name}.json'
        with open(dashboard_path, 'w') as file:
            json.dump(dashboard_manifest, file)

        # Log or handle the fact that files are written
        logger.info(f"Datasource files written: {data_source_path}")
        logger.info(f"Dashboard files written: {dashboard_path}")
    def setup_routes(self):
        @self.app.get("/jobs")
        def get_jobs():
            # Directly return the timekeeper jobs data as JSON
            return self.timekeeper.get_jobs()

        @self.app.get("/archive")
        def get_archive():
            # Assuming Timekeeper has a method to get archive data
            return self.timekeeper.get_archive()

    def start(self):
        self.logger.info("Starting DataSourceService...")
        self.server.start()
        self.logger.info("DataSourceService started.")

    def stop(self):
        self.logger.info("Stopping DataSourceService...")
        self.server.stop()
        self.logger.info("DataSourceService stopped.")


from datetime import datetime, timedelta
from enum import Enum

from sonaris.device.dg4202 import DG4202
from sonaris.defaults import DeviceName
from sonaris import factory
from sonaris.frontend.tasks.task_decorator import (
    parameter_annotations,
    parameter_constraints,
)

"""
These tasks are used by the scheduler, they are wrappers for the scheduler to call the manager objects.
You will have to point to them under header.py
"""


class TaskName(Enum):
    DG4202_TOGGLE = "Toggle Output"
    DG4202_SET_WAVEFORM = "Set Waveform Parameters"
    DG4202_SET_SWEEP = "Set Sweep Parameters"
    EDUX1002A_AUTO = "Press Auto"


@parameter_constraints(channel=(1, 2), output=["ON", "OFF"])
def task_on_off_dg4202(channel: int, output: bool) -> bool:
    factory.dg4202_manager.device.output_on_off(
        channel=channel,
        status=output,  # the decorator above will be handled by ui_factory.py
    )
    return True


@parameter_annotations(
    frequency="Hz",
    offset="V",
    amplitude="V",
)
@parameter_constraints(
    frequency=(0.0, DG4202.FREQ_LIMIT),
    channel=[1, 2],  # decorating with list means forced options
    waveform_type=DG4202.available_waveforms(),
    offset=(0.0, 5.0),
    amplitude=(0.0, 5.0),
)
def task_set_waveform_parameters(
    channel: int,
    send_on: bool,
    waveform_type: str,
    amplitude: float,
    frequency: float,
    offset: float,
) -> bool:
    factory.dg4202_manager.device.set_waveform(
        channel=channel,
        waveform_type=waveform_type,
        amplitude=amplitude,
        frequency=frequency,
        params=None,
        offset=offset,
    )
    if send_on:
        factory.dg4202_manager.device.output_on_off(channel, True)
    return True


@parameter_annotations(
    fstart="Hz",
    fstop="Hz",
    time="s",
    rtime="ms",
    htime_start="ms",
    htime_stop="ms",
)
@parameter_constraints(
    channel=[1, 2],
    fstart=(0.0, DG4202.FREQ_LIMIT),
    fstop=(0.0, DG4202.FREQ_LIMIT),
    time=(0.0, float("inf")),
    rtime=(0.0, float("inf")),
    htime_start=(0.0, float("inf")),
    htime_stop=(0.0, float("inf")),
)
def task_set_sweep_parameters(
    channel: int,
    send_on: bool,
    fstart: float,
    fstop: float,
    time: float,
    rtime: float = 0,
    htime_start: float = 0,
    htime_stop: float = 0,
) -> bool:
    params = {
        "FSTART": fstart,
        "FSTOP": fstop,
        "TIME": time,
        "RTIME": rtime,
        "HTIME_START": htime_start,
        "HTIME_STOP": htime_stop,
    }
    factory.dg4202_manager.device.set_sweep_parameters(
        channel=channel, sweep_params=params
    )
    if send_on:
        factory.dg4202_manager.device.output_on_off(channel, True)
    return True


@parameter_constraints(press=["OK"])
def task_auto_edux1002a(press: str):
    # for testing, kwarg_value means nothing
    factory.edux1002a_manager.device.autoscale()
    return True


"""
The task list is read by the job_scheduler module and app.py to register the task and render the UI.
"""

TASK_LIST_DICTIONARY = {
    DeviceName.DG4202.value: {
        TaskName.DG4202_TOGGLE.value: task_on_off_dg4202,
        TaskName.DG4202_SET_WAVEFORM.value: task_set_waveform_parameters,
        TaskName.DG4202_SET_SWEEP.value: task_set_sweep_parameters,
    },
    DeviceName.EDUX1002A.value: {TaskName.EDUX1002A_AUTO.value: task_auto_edux1002a},
}


def get_tasks(flatten: bool = False) -> dict:
    """Returns the dict of { device : { task-name : func_pointer , ..} ..}

    Returns:
        dict: dictionary containing devices and its tasks.
    """
    if flatten:
        return {
            inner_key: value
            for outer_dict in TASK_LIST_DICTIONARY.values()
            for inner_key, value in outer_dict.items()
        }
    return TASK_LIST_DICTIONARY


def calculate_schedule_times(steps):
    """
    Calculate the schedule times for each step based on wait and at_time keys.
    :param steps: A list of step dictionaries from the experiment configuration.
    :return: A list of tuples (step, schedule_time) with calculated schedule times.
    """
    last_schedule_time = datetime.now()  # Initialize with the current time
    schedule_times = []

    for step in steps:
        if "at_time" in step:
            at_time = timedelta(seconds=step["at_time"])
            schedule_time = datetime.now() + at_time
        else:
            wait_time = timedelta(seconds=step.get("wait", 0))
            schedule_time = last_schedule_time + wait_time

        schedule_times.append((step, schedule_time))
        last_schedule_time = schedule_time  # Update for the next iteration

    return schedule_times




from sonaris.utils.container import start_container, stop_container, remove_container, client,network
from typing import NoReturn, Optional
import hashlib
from docker.models.containers import Container
from docker import DockerClient
from sonaris.defaults import APP_NAME,SONARIS_NETWORK_NAME, GF_SECURITY_ADMIN_PASSWORD, GF_PORT, GF_DASHBOARDS_DIR,GF_DATASOURCES_DIR
from sonaris.utils.log import get_logger
from sonaris.services.service import Service
from pathlib import Path
from docker.models.networks import Network 
logger= get_logger()


class GrafanaContainerService(Service):
    def __init__(self, client: DockerClient = None ,port:int = None,datasources_dir:Path = None, dashboards_dir:Path = None,network_name: str = SONARIS_NETWORK_NAME) -> None:
        self._client :DockerClient = client
        self._network_name :str= network_name
        self.ensure_docker_client()
        self._network :Network = None
        self.ensure_docker_network()
        self.port :int= port or GF_PORT
        self.container_label = self.get_container_label()
        self.datasources_dir :Path = datasources_dir or GF_DATASOURCES_DIR
        self.dashboards_dir :Path= dashboards_dir or GF_DASHBOARDS_DIR
    def ensure_docker_client(self):
        """Ensure the Docker client is available and connected."""
        if self._client is None:
            try:
                self._client = DockerClient.from_env()
                logger.info(f"Docker client obtained: {self._client}")
                self._client.ping()  # Check connection to Docker
            except Exception as e:
                logger.error(f"Failed to obtain Docker client: {e}")
                self._client = None

    def ensure_docker_network(self):
        """Ensure the Docker network is available."""
        if self._network is None and self._client is not None:
            try:
                self._network = self._client.networks.get(self._network_name)
                logger.info(f"Found existing network: {self._network_name}")
            except Exception as e:
                try:
                    self._network = self._client.networks.create(self._network_name, driver="bridge", check_duplicate=True)
                    logger.info(f"Created new network: {self._network_name}")
                except Exception as e:
                    logger.error(f"Failed to create or retrieve network: {e}")
                    self._network = None

    @property
    def client(self):
        """Property to get the Docker client, fetching it if necessary."""
        if self._client is None:
            self.ensure_docker_client()
        return self._client

    @property
    def network(self):
        """Property to get the Docker network, creating or fetching it if necessary."""
        if self._network is None:
            self.ensure_docker_network()
        return self._network
    def _generate_hash_(self) -> str:
        """Static hash for container labeling."""
        return hashlib.sha256(str(APP_NAME).encode()).hexdigest()[:10]

    def find_grafana_container(self) -> Optional[Container]:
        """
        Find the Grafana container by label.
        :return: The Grafana Container if found, None otherwise.
        """
        if self.client is None:
            return None
        containers = self.client.containers.list(all=True, filters={"label": f"{APP_NAME.lower()}={self.container_label}"})
        return containers[0] if containers else None
    
    def get_container_label(self) -> str:
        return f"{APP_NAME.lower()}_{self._generate_hash_()}"
    
    def is_docker_client_online(self) -> bool:
        if self.client is None:
            logger.error("Docker client is not found. start Docker and restart the application.")
            return False
        return 
    
    def create_grafana_container(self, image: str = "grafana/grafana", network_name: str = "sonaris_network") -> Optional[str]:
        """Create and start a Grafana container on a custom Docker network."""
        if self.client is None:
            logger.error("Docker client is unavailable. Cannot start Grafana service.")
            return None
        try:
            container_label = self.get_container_label()
            container: Container = self.client.containers.run(
                image,
                ports={'3000/tcp': self.port},  # Map Grafana's default port to the specified port on the host
                environment={
                    'GF_SECURITY_ADMIN_PASSWORD': f'{GF_SECURITY_ADMIN_PASSWORD}',
                },
                labels={f"{APP_NAME.lower()}": container_label},
                detach=True
            )
            if self.network:
                self.network.connect(container)
            logger.info(f"Grafana container {container_label} created and started on custom network '{network_name}'. Accessible on http://localhost:{self.port}.")
            return container.id
        except Exception as e:
            logger.error(f"Failed to create Grafana container: {e}")
            return None
    def start(self) -> None:
        """
        Start the Grafana service by finding or creating the necessary container.
        """
        if self.client is None:
            logger.error("Docker client is unavailable. Cannot start Grafana service.")
            return

        container = self.find_grafana_container()
        if container is None:
            logger.info("Grafana container not found, creating a new one...")
            self.create_grafana_container()
        else:
            logger.info(f"Found existing Grafana container with container id: {container.id}")
            if container.status != 'running':
                container.start()
                logger.info("Grafana container started.")
            else:
                logger.info("Grafana container is already running.")

    def stop(self) -> None:
        """
        Stop the Grafana service.
        """
        if self.client is None:
            logger.error("Docker client is unavailable. Cannot stop Grafana service.")
            return

        container = self.find_grafana_container()
        if container and container.status == 'running':
            container.stop()
            logger.info("Grafana container stopped.")
        else:
            logger.info("No running Grafana container found to stop.")
            
    def stop_container(self, container_id: str) -> NoReturn:
        stop_container(container_id)
        
    def start_container(self, container_id: str) -> NoReturn:
        start_container(container_id)

    def recreate_grafana_container_with_volumes(self):
        if self.client is None:
            logger.error("Docker client is unavailable. Cannot recreate Grafana service.")
            return 
        existing_container = self.find_grafana_container()
        if existing_container:
            existing_container.stop()
            existing_container.remove()
        self.create_grafana_container()